/*
 * Created By Drake Vidkjer
 * Created on 10-05-17
 * Modified on 10-10-17
 * Simple program to display processor information using the CPUID assembly call
 */

#include <stdio.h>
#include <string.h>
#include "lookup.h"

/*
 * Call the cpuid instruction with the appropriate value in eax. values of variables from main function
 * directly accessed throuch pointers
 */
void callCPUID(unsigned int initEAX, unsigned int *eax, unsigned int *ebx, unsigned int *ecx, unsigned int *edx)
{
	asm volatile 
	("cpuid"
		: "=a"(*eax), "=b"(*ebx), "=c"(*ecx), "=d"(*edx)
		: "a"(initEAX)
	);
	return;
}

int main(int argc, char **argv)
{
	//print out some information about the program
	printf("*******************************************************\n");
	printf("ECE 472 HW1, CPUID Information\nCreated by Drake Vidkjer on");
	printf(" 10-05-17\nModified on 10-10-17\nThis program reports");
	printf(" information on the installed processor.");
	printf("\n*****************************************************\n");
	
	
	//define general variables
	unsigned int a, b, c, d, j;
	
	//Print out vendor info
	char vendor[13];
	callCPUID(0x0, &a, &b, &c, &d);
	memcpy(vendor, (char*)&b, 4);
	memcpy(vendor + 4, (char*)&d, 4);
	memcpy(vendor + 8, (char*)&c, 4);
	vendor[12] = 0;
	printf("Vendor: %s\n", vendor);
	
	//Print CPU family
	callCPUID(0x1, &a, &b, &c, &d);
	printf("Stepping %d\n", a & 0xF);
	printf("Model %d\n", (a >> 4) & 0xF);
	printf("Extended Model %d\n", (a >> 16) & 0xF);
	printf("Family %d\n", (a >> 8) & 0xF);
	printf("Extended Family %d\n", (a >> 20) & 0xFF);
	
	//Print CPU model (number and name)
	//I dont know where I could get this information, I know from the model 
	//number I can get it but there doesn't seem to be any sort of clean way to
	//do it and also all the tables I found were out of date and ended with the
	//core 2 family.
	
	//Print number of logical CPUs
	c = 0x0;
	callCPUID(0xB, &a, &b, &c, &d);
	printf("\nLogical CPUs: %d\n", (b >> 16) & ~(~0 << 8));
	printf("\nLogical CPUs: %d\n", b & 0xF);
	
	//Print CPU frequency
	callCPUID(0x16, &a, &b, &c, &d);
	//printf("\nProcessor Base Frequency: %dMHz", (a >> 16) & 0xF);
	printf("\nProcessor Base Frequency: %dMHz", a & 0xF);
	printf("\nProcessor Max Frequency: %dMHz", b);
	printf("\nBus Reference Frequency: %dMHz\n", c);
	
	//Print CPU features available
	callCPUID(0x1, &a, &b, &c, &d);
	printf("\nSupported Features:");
	//First check edx then ecx
	for(j = 0; j < 32; j++)
	{
		if(d & (1 << j))
		{
			printf(" %s", featureEDX[j]);
		}
	}
	for(j = 0; j < 32; j++)
	{
		if(c & (1 << j))
		{
			printf(" %s", featureECX[j]);
		}
	}
	
	 //Debug code
	 //printf("\nA:%#010x\n", a);
	 //printf("B:%#010x\n", b);
	 //printf("C:%#010x\n", c);
	 //printf("D:%#010x\n", d);
	
	/*
	 * Print L1 data cache size
	 * Print L1 instruction cache size
	 * Print TLB info
	 * To do this, I'll call the function with eax=4; then go through
	 * the different bits of each register and print the appropriate
	 * entry from the array. (This doesnt work because the array of 
	 * TLB / Cache info does 
	 */
	 callCPUID(0x2, &a, &b, &c, &d);
	 printf("\n\nCache and TLB information (Must lookup in Intel documentation):\n");
	 if(((a >> 31) & 0x01) == 0)
	 {
		printf("%#04x\n", (a >> 8) & 0xFF);
		printf("%#04x\n", (a >> 16) & 0xFF);
		printf("%#04x\n", (a >> 24) & 0xFF);
	 }
	 //printf("%s\n", cacheInfo[(a) & 0xFF]);
	 if(((b >> 31) & 0x01) == 0)
	 {
		printf("%#04x\n", (b) & 0xFF);
		printf("%#04x\n", (b >> 8) & 0xFF);
		printf("%#04x\n", (b >> 16) & 0xFF);
		printf("%#04x\n", (b >> 24) & 0xFF);
	 }
	 if(((c >> 31) & 0x01) == 0)
	 {
		printf("%#04x\n", (c) & 0xFF);
		printf("%#04x\n", (c >> 8) & 0xFF);
		printf("%#04x\n", (c >> 16) & 0xFF);
		printf("%#04x\n", (c >> 24) & 0xFF);
	 }
	 if(((d >> 31) & 0x01) == 0)
	 {
		printf("%#04x\n", (d) & 0xFF);
		printf("%#04x\n", (d >> 8) & 0xFF);
		printf("%#04x\n", (d >> 16) & 0xFF);
		printf("%#04x\n", (d >> 24) & 0xFF);
	 }
	 
	 //here's another way to get cache size
	 callCPUID(0x80000006, &a, &b, &c, &d);
	 printf("\nL1 Cache size in bytes: %d", (c & 0xFF));
	 printf("\nL1 Data Cache size in bytes: %d", (c & 0xFF)/2);
	 printf("\nL1 Instruction Cache size in bytes: %d", (c & 0xFF)/2);
	 
	//Print Memory hierarchy information
	//I couldnt find any information on how to get memory hierarchy information from the cpuid
	//system call
	
	//Print physical address width
	//Print logical address width
	//call the cpuid with the correct eax, and mask off the correct values in eax
	callCPUID(0x80000008, &a, &b, &c, &d);
	printf("\nPhysical Address Size: %d", a & 0xF);
	printf("\nLogical Address Size: %d", (a >> 8) & 0xF);
	
	printf("\n");
	 
	return 0;
}
